/*
 * Copyright (c) 2012 Marsha Chechik, Alessio Di Sandro, Michalis Famelis,
 * Rick Salay, Vivien Suen.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    Wentao He - Implementation.
 */
package edu.toronto.cs.se.modelepedia.erdiagram.operator;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;

import edu.toronto.cs.se.mmtf.mid.Model;
import edu.toronto.cs.se.mmtf.mid.ModelOrigin;
import edu.toronto.cs.se.mmtf.mid.MultiModel;
import edu.toronto.cs.se.mmtf.mid.operator.impl.OperatorExecutableImpl;
import edu.toronto.cs.se.mmtf.mid.relationship.ModelElementEndpoint;
import edu.toronto.cs.se.mmtf.mid.relationship.ModelRel;
import edu.toronto.cs.se.mmtf.mid.trait.MultiModelInstanceFactory;
import edu.toronto.cs.se.mmtf.mid.trait.MultiModelOperatorUtils;
import edu.toronto.cs.se.mmtf.mid.trait.MultiModelRegistry;
import edu.toronto.cs.se.mmtf.mid.trait.MultiModelTypeIntrospection;
import edu.toronto.cs.se.modelepedia.erdiagram.Attribute;
import edu.toronto.cs.se.modelepedia.erdiagram.ERDiagram;
import edu.toronto.cs.se.modelepedia.erdiagram.Entity;
import edu.toronto.cs.se.modelepedia.erdiagram.ErdiagramFactory;
import edu.toronto.cs.se.modelepedia.erdiagram.ErdiagramPackage;
import edu.toronto.cs.se.modelepedia.erdiagram.Relationship;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.enumeration.LevelEnum;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.enumeration.SourceEnum;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.util.ERDiagramMergeUtil;

/**
 * An implementation of ER diagram merge.
 */
public class ERDiagramMerge extends OperatorExecutableImpl {

	private static final String FILE_SUFFIX = "_hwt";
	private static final String MERGED_MODEL_NAME = "MergedDiagram";
	
	private static final String ENTITY_PREFIX = "ENTITY@";
	private static final String RELATIONSHIP_PREFIX = "RELATIONSHIP@";
	private static final String ENTITY_ATTRIBUTE_PREFIX = "ENTITY_ATTRIBUTE@";
	private static final String RELATIONSHIP_ATTRIBUTE_PREFIX = "RELATIONSHIP_ATTRIBUTE@";
	private static final String ENTITY = "Entity";
	private static final String RELATIONSHIP = "Relationship";
	private static final String ATTRIBUTE = "Attribute";
	private static final String LEFT_SUFFIX = "@Left";
	private static final String RIGHT_SUFFIX = "@Right";
	private static final String CONNECTOR_SUFFIX = "@Connector";
	
	private static final String PROPERTY_IN_ISOPENWORLD = "isOpenWorld";
	private static final String PROPERTY_IN_PREFERRIGHTMODELNAMING = "preferRightModelNaming";
	
	private static final String DASH = "-";
	private static final String DOT = ".";
	private static final String SPLIT = "^";
	private static final String AT = "@";
	private static final String BLANK = "";
	
	ErdiagramFactory erdiagramFactory = ErdiagramFactory.eINSTANCE;
	List<Entity> entityDisjointUnion;
	List<Relationship> relationshipDisjointUnion;
	List<Attribute> attributeDisjointUnion;
	
	SourceEnum namingPriority = SourceEnum.RIGHT_DIAGRAM;
	boolean isOpenWorld;
	boolean preferRightModelNaming;
	
	@Override
	public EList<Model> execute(EList<Model> actualParameters) throws Exception {
		EList<Model> result = new BasicEList<Model>();
		Model mergedERDiagramModel = null; 
		
		ERDiagram leftERDiagram = (ERDiagram)MultiModelTypeIntrospection.getRoot(actualParameters.get(0));
		ERDiagram rightERDiagram = (ERDiagram)MultiModelTypeIntrospection.getRoot(actualParameters.get(1));
		
		try{
			Properties inputProperties = MultiModelOperatorUtils.getPropertiesFile(
					this,
					actualParameters.get(0),
					null,
					MultiModelOperatorUtils.INPUT_PROPERTIES_SUFFIX
				);
			readProperties(inputProperties);	
		} catch(Exception e) {
			//Error happened while reading the properties file specified by the end user. Use default settings.
			isOpenWorld = false;
			preferRightModelNaming = true;
		}
		
		Map<String, String> leftToRightDiagramMapping = validateAndPopulateInputDiagramMapping((ModelRel) actualParameters.get(2), 
				leftERDiagram, rightERDiagram);
		
		if(!preferRightModelNaming) {
			ERDiagram tempERDiagram = leftERDiagram;
			leftERDiagram = rightERDiagram;
			rightERDiagram = tempERDiagram;
			
			Map<String, String> tempMapping = new HashMap<String, String>();
			for(int i=0; i<leftToRightDiagramMapping.size(); i++) {
				String key = (String)leftToRightDiagramMapping.keySet().toArray()[i];
				tempMapping.put(leftToRightDiagramMapping.get(key), key);
			}
			leftToRightDiagramMapping = tempMapping;
		}
		
		ERDiagram mergedERDiagram = merge(leftERDiagram, rightERDiagram, leftToRightDiagramMapping);
			
		//Serialize
		String mergedERDiagramModelUri = MultiModelRegistry.replaceFileExtensionInUri(actualParameters.get(0).getUri(), ErdiagramPackage.eNAME);
		mergedERDiagramModelUri = MultiModelRegistry.addFileNameSuffixInUri(mergedERDiagramModelUri, FILE_SUFFIX + (new Date()).getTime());
		MultiModelTypeIntrospection.writeRoot(mergedERDiagram, mergedERDiagramModelUri, true);
		
		//Create Model
		MultiModel multiModel = MultiModelRegistry.getMultiModel(actualParameters.get(0));
		mergedERDiagramModel = MultiModelInstanceFactory.createModelAndEditor(null, mergedERDiagramModelUri, ModelOrigin.CREATED, multiModel);
		mergedERDiagramModel.setName(MERGED_MODEL_NAME);
		
		result.add(mergedERDiagramModel);
		
		return result;
	}
	
	
	/**
	 * Merge two input ER diagrams.
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @param leftToRightDiagramMapping the mapping between the two input ER diagrams
	 * @return merged ER diagram
	 */
	private ERDiagram merge(ERDiagram leftERDiagram, ERDiagram rightERDiagram, Map<String, String> leftToRightDiagramMapping) {		
		ERDiagram mergedERDiagram = erdiagramFactory.createERDiagram();
		ERDiagram connector = createConnector(rightERDiagram, leftToRightDiagramMapping); 
		
		Map<String, String> connectorToLeftDiagramMapping = createConnectorToDiagramMapping(leftToRightDiagramMapping, SourceEnum.LEFT_DIAGRAM);
		Map<String, String> connectorToRightDiagramMapping = createConnectorToDiagramMapping(leftToRightDiagramMapping, SourceEnum.RIGHT_DIAGRAM);
		
		//start merging set	
		List<String> entityDisjointUnion = populateDisjointUnion(leftERDiagram, rightERDiagram, connector, LevelEnum.ENTITY);
		List<String> relationshipDisjointUnion = populateDisjointUnion(leftERDiagram, rightERDiagram, connector, LevelEnum.RELTIONSHIP);
		List<String> entityAttributeDisjointUnion = populateDisjointUnion(leftERDiagram, rightERDiagram, connector, LevelEnum.ATTRIBUTE_OF_ENTITY);
		List<String> relationshipAttributeDisjointUnion = populateDisjointUnion(leftERDiagram, rightERDiagram, connector, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP);
		
		List<List<String>> mergedEntitySet = mergeSet(ERDiagramMergeUtil.getEntityMapping(connectorToLeftDiagramMapping), 
				ERDiagramMergeUtil.getEntityMapping(connectorToRightDiagramMapping), entityDisjointUnion, LevelEnum.ENTITY);
		List<List<String>> mergedRelationshipSet = mergeSet(ERDiagramMergeUtil.getRelationshipMapping(connectorToLeftDiagramMapping), 
				ERDiagramMergeUtil.getRelationshipMapping(connectorToRightDiagramMapping), relationshipDisjointUnion, LevelEnum.RELTIONSHIP);	
		List<List<String>> mergedEntityAttributeSet = mergeSet(ERDiagramMergeUtil.getEntityAttributeMapping(connectorToLeftDiagramMapping), 
				ERDiagramMergeUtil.getEntityAttributeMapping(connectorToRightDiagramMapping), entityAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_ENTITY);
		List<List<String>> mergedRelationshipAttributeSet = mergeSet(ERDiagramMergeUtil.getRelationshipAttributeMapping(connectorToLeftDiagramMapping), 
				ERDiagramMergeUtil.getRelationshipAttributeMapping(connectorToRightDiagramMapping), 
				relationshipAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP);
		
		List<String> distinctMergedEntitySet = yieldDistinctMergedSet(mergedEntitySet);
		List<String> distinctMergedRelationshipSet = yieldDistinctMergedSet(mergedRelationshipSet);
		List<String> distinctMergedEntityAttributeSet = yieldDistinctMergedSet(mergedEntityAttributeSet);
		List<String> distinctMergedRelationshipAttributeSet = yieldDistinctMergedSet(mergedRelationshipAttributeSet);
		
		//start merging graph
		mergeGraphWithEntities(distinctMergedEntitySet, leftERDiagram, rightERDiagram, connector, mergedERDiagram);
		mergeGraphWithRelationships(distinctMergedRelationshipSet, mergedEntitySet, distinctMergedEntitySet, 
				leftERDiagram, rightERDiagram, connector, mergedERDiagram);
		mergeGraphWithEntityAttributes(distinctMergedEntityAttributeSet, mergedEntityAttributeSet, 
				leftERDiagram, rightERDiagram, connector, mergedERDiagram);
		mergeGraphWithRelationshipAttributes(distinctMergedRelationshipAttributeSet, mergedRelationshipAttributeSet, 
				leftERDiagram, rightERDiagram, connector, mergedERDiagram);

		return mergedERDiagram;
	}
	
	
	/**
	 * Create connector for the two input ER diagrams based on the input ER diagram and the mapping.
	 * @param rightERDiagram the left side input ER diagram
	 * @param leftToRightDiagramMapping the mapping between the two input ER diagrams
	 * @return connector for the two input ER diagrams
	 */
	private ERDiagram createConnector(ERDiagram rightERDiagram, Map<String, String> leftToRightDiagramMapping) {
		ERDiagram connector = erdiagramFactory.createERDiagram();
		Map<String, String> entityMapping = new HashMap<String, String>();
		Map<String, String> relationshipMapping = new HashMap<String, String>();
		Map<String, String> entityAttributeMapping = new HashMap<String, String>();
		Map<String, String> relationshipAttributeMapping = new HashMap<String, String>();
		
		for(int i=0; i<leftToRightDiagramMapping.size(); i++) {
			String key = (String)leftToRightDiagramMapping.keySet().toArray()[i];
			if(key.startsWith(ENTITY_PREFIX))
				entityMapping.put(key, leftToRightDiagramMapping.get(key));
			if(key.startsWith(RELATIONSHIP_PREFIX))
				relationshipMapping.put(key, leftToRightDiagramMapping.get(key));
			if(key.startsWith(ENTITY_ATTRIBUTE_PREFIX))
				entityAttributeMapping.put(key, leftToRightDiagramMapping.get(key));
			if(key.startsWith(RELATIONSHIP_ATTRIBUTE_PREFIX))
				relationshipAttributeMapping.put(key, leftToRightDiagramMapping.get(key));
		}
		
		for(int i=0; i<entityMapping.size(); i++) {
			String value = (String)entityMapping.values().toArray()[i];
			String entityName = value.substring(ENTITY_PREFIX.length());
			Entity entity = ERDiagramMergeUtil.lookupEntity(rightERDiagram, entityName);
			Entity clonedEntity = ERDiagramMergeUtil.cloneEntity(entity);
			connector.getEntities().add(clonedEntity);	
		}
		
		for(int i=0; i<relationshipMapping.size(); i++) {
			String value = (String)relationshipMapping.values().toArray()[i];
			String relationshipName = value.substring(RELATIONSHIP_PREFIX.length());
			Relationship relationship = ERDiagramMergeUtil.lookupRelationship(rightERDiagram, relationshipName);
			Relationship clonedRelationship = ERDiagramMergeUtil.cloneRelationship(relationship);
			clonedRelationship.setFrom(ERDiagramMergeUtil.lookupEntity(connector, relationship.getFrom().getName()));
			clonedRelationship.setTo(ERDiagramMergeUtil.lookupEntity(connector, relationship.getTo().getName()));
			connector.getRelationships().add(clonedRelationship);
		}
		
		for(int i=0; i<entityAttributeMapping.size(); i++) {
			String value = (String)entityAttributeMapping.values().toArray()[i];
			String entityAttributeName = value.substring(ENTITY_ATTRIBUTE_PREFIX.length());
			Attribute entityAttribute = ERDiagramMergeUtil.lookupEntityAttribute(rightERDiagram, entityAttributeName);
			Attribute clonedEntityAttribute = ERDiagramMergeUtil.cloneAttribute(entityAttribute);
			Entity entity = ERDiagramMergeUtil.lookupEntity(connector, entityAttributeName.substring(0, entityAttributeName.lastIndexOf(DASH)));
			entity.getEntityAttributes().add(clonedEntityAttribute);
		}
		
		for(int i=0; i<relationshipAttributeMapping.size(); i++) {
			String value = (String)relationshipAttributeMapping.values().toArray()[i];
			String relationshipAttributeName = value.substring(RELATIONSHIP_ATTRIBUTE_PREFIX.length());
			Attribute relationshipAttribute = ERDiagramMergeUtil.lookupRelationshipAttribute(rightERDiagram, relationshipAttributeName);
			Attribute clonedRelationshipAttribute = ERDiagramMergeUtil.cloneAttribute(relationshipAttribute);
			Relationship relationship = ERDiagramMergeUtil.lookupRelationship(connector, 
					relationshipAttributeName.substring(0, relationshipAttributeName.lastIndexOf(DASH)));
			relationship.getRelationshipAttributes().add(clonedRelationshipAttribute);
		}
		
		return connector;
	}

	
	/**
	 * Create mapping between connector to each a specified input ER diagram.
	 * @param leftToRightDiagramMapping the mapping between the two input ER diagrams
	 * @param source an indication of which input ER diagram to create mapping for
	 * @return mapping between connector to the specified input ER diagram
	 */
	private Map<String, String> createConnectorToDiagramMapping(Map<String, String> leftToRightDiagramMapping, SourceEnum source) {
		Map<String, String> connectorToDiagramMapping = new HashMap<String, String>();
		
		if(SourceEnum.RIGHT_DIAGRAM.equals(source)) {
			for(int i=0; i<leftToRightDiagramMapping.size(); i++) {
				String value = (String)leftToRightDiagramMapping.values().toArray()[i];
				connectorToDiagramMapping.put(value, value);
			}			
		} else if(SourceEnum.LEFT_DIAGRAM.equals(source)) {
			for(int i=0; i<leftToRightDiagramMapping.size(); i++) {
				String key = (String)leftToRightDiagramMapping.keySet().toArray()[i];
				String value = (String)leftToRightDiagramMapping.values().toArray()[i];
				connectorToDiagramMapping.put(value, key);
			}	
		}
			
		return connectorToDiagramMapping;
	}
	
	
	/**
	 * Populate disjoint union for the two input ER diagrams of specified element level.
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @param connector the connector ER diagram for the two input ER diagrams
	 * @param level the element level of populated disjoint union
	 * @return disjoint union for the two input ER diagrams of specified element level
	 */
	private List<String> populateDisjointUnion(ERDiagram leftERDiagram, ERDiagram rightERDiagram, ERDiagram connector, LevelEnum level) {
		List<String> disjointUnion = new ArrayList<String>();
		
		if(LevelEnum.ENTITY.equals(level)) {
			for(Entity entity: leftERDiagram.getEntities()) 
				disjointUnion.add(entity.getName() + LEFT_SUFFIX);
			for(Entity entity: rightERDiagram.getEntities()) 
				disjointUnion.add(entity.getName() + RIGHT_SUFFIX);
			for(Entity entity: connector.getEntities())
				disjointUnion.add(entity.getName()+ CONNECTOR_SUFFIX);
		}		
		else if(LevelEnum.RELTIONSHIP.equals(level)) {
			for(Relationship relationship: leftERDiagram.getRelationships())
				disjointUnion.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName() + LEFT_SUFFIX);
			for(Relationship relationship: rightERDiagram.getRelationships())
				disjointUnion.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName() + RIGHT_SUFFIX);
			for(Relationship relationship: connector.getRelationships())
				disjointUnion.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName() + CONNECTOR_SUFFIX);	
		}
		else if(LevelEnum.ATTRIBUTE_OF_ENTITY.equals(level)) {
			addEntityAttributes(disjointUnion, leftERDiagram, LEFT_SUFFIX);
			addEntityAttributes(disjointUnion, rightERDiagram, RIGHT_SUFFIX);
			addEntityAttributes(disjointUnion, connector, CONNECTOR_SUFFIX);			
		}
		else if(LevelEnum.ATTRIBUTE_OF_RELATIONSHIP.equals(level)) {
			addRelationshipAttributes(disjointUnion, leftERDiagram, LEFT_SUFFIX);
			addRelationshipAttributes(disjointUnion, rightERDiagram, RIGHT_SUFFIX);
			addRelationshipAttributes(disjointUnion, connector, CONNECTOR_SUFFIX);
		}
		
		return disjointUnion;
	}
	
	
	/**
	 * Add entity attributes to the disjoint union.
	 * @param disjointUnion the current disjoint union
	 * @param erdiagram
	 * @param suffix
	 */
	private void addEntityAttributes(List<String> disjointUnion, ERDiagram erdiagram, String suffix) {
		for(Entity entity: erdiagram.getEntities()) {
			for(Attribute attribute: entity.getEntityAttributes()) {
				disjointUnion.add(entity.getName() + DASH + attribute.getName() + suffix);
				if(attribute.getAttributeAttributes().size() != 0)
					for(Attribute subAttribute: attribute.getAttributeAttributes())
						disjointUnion.add(entity.getName() + DASH + attribute.getName() + SPLIT + subAttribute.getName() + suffix);
			}
		}
	}
	
	
	/**
	 * Add relationship attributes to the disjoint union.
	 * @param disjointUnion the current disjoint union
	 * @param erdiagram
	 * @param suffix
	 */
	private void addRelationshipAttributes(List<String> disjointUnion, ERDiagram erdiagram, String suffix) {
		for(Relationship relationship: erdiagram.getRelationships()) {
			for(Attribute attribute: relationship.getRelationshipAttributes()) {
				disjointUnion.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName()
						+ DASH + attribute.getName() + suffix);
			}
		}
	}
	
	
	/**
	 * Create merged set, the merged set is a list of list, each list contains mapped elements from left ER diagram, right ER diagram 
	 * and connector (if applicable).
	 * @param connectorToLeftMapping the mapping between connector to left side ER diagram
	 * @param connectorToRightMapping the mapping between connector to right side ER diagram
	 * @param disjointUnion the disjointUnion of the two input ER diagrams
	 * @param level the element level
	 * @return the merged set for the specified element level
	 */
	private List<List<String>> mergeSet(Map<String, String> connectorToLeftMapping, Map<String, String> connectorToRightMapping, 
			List<String> disjointUnion, LevelEnum level) {
		List<List<String>> mergedSet = ERDiagramMergeUtil.convertToListOfList(disjointUnion);
		int connectorToLeftMappingSize = connectorToLeftMapping.size();
		int connectorToRightMappingSize = connectorToRightMapping.size();
		String from = BLANK;
		String to = BLANK;
				
		for(int i=0; i<connectorToLeftMappingSize+connectorToRightMappingSize; i++) {
			String key = i<connectorToLeftMappingSize ? (String)connectorToLeftMapping.keySet().toArray()[i] 
					: (String)connectorToRightMapping.keySet().toArray()[i-connectorToRightMappingSize];
			String value = i<connectorToLeftMappingSize ? (String)connectorToLeftMapping.values().toArray()[i] 
					: (String)connectorToRightMapping.values().toArray()[i-connectorToRightMappingSize];
			
			if(LevelEnum.ENTITY.equals(level)) {
				from = key.substring(ENTITY_PREFIX.length()) + CONNECTOR_SUFFIX;
				to = value.substring(ENTITY_PREFIX.length()) + (i<connectorToLeftMappingSize ? LEFT_SUFFIX : RIGHT_SUFFIX);
			}
			else if(LevelEnum.RELTIONSHIP.equals(level)) {
				from = key.substring(RELATIONSHIP_PREFIX.length()) + CONNECTOR_SUFFIX;
				to = value.substring(RELATIONSHIP_PREFIX.length()) + (i<connectorToLeftMappingSize ? LEFT_SUFFIX : RIGHT_SUFFIX);
			}
			else if(LevelEnum.ATTRIBUTE_OF_ENTITY.equals(level)) {
				from = key.substring(ENTITY_ATTRIBUTE_PREFIX.length()) + CONNECTOR_SUFFIX;
				to = value.substring(ENTITY_ATTRIBUTE_PREFIX.length()) + (i<connectorToLeftMappingSize ? LEFT_SUFFIX : RIGHT_SUFFIX);
			}
			else if(LevelEnum.ATTRIBUTE_OF_RELATIONSHIP.equals(level)) {
				from = key.substring(RELATIONSHIP_ATTRIBUTE_PREFIX.length()) + CONNECTOR_SUFFIX;
				to = value.substring(RELATIONSHIP_ATTRIBUTE_PREFIX.length()) + (i<connectorToLeftMappingSize ? LEFT_SUFFIX : RIGHT_SUFFIX);
			}
			
			ERDiagramMergeUtil.mergeList(mergedSet, from, to);
		}	
		
		return mergedSet;
	}
	
	
	/**
	 * Take one element from each list from merged set, yield another list which contains unique elements. 
	 * @param mergedSet the merged set of a specified element level
	 * @return list containing unique merged elements
	 */
	private List<String> yieldDistinctMergedSet(List<List<String>> mergedSet) {
		List<String> distinctMergedSet = new ArrayList<String>();
		
		for(List<String> list: mergedSet) {
			if(list.size() == 1) {
				distinctMergedSet.add(list.get(0));
			} else if(list.size() > 1) {
				for(String item: list)
					if(item.endsWith(CONNECTOR_SUFFIX))
						distinctMergedSet.add(item);
			}
		}
		
		return distinctMergedSet;
	}
	
	
	/**
	 * Construct the final graph, add all merged entities to the graph.
	 * @param distinctMergedEntitySet the list containing unique merged elements
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @param connector the connector ER diagram for the two input ER diagrams
	 * @param mergedERDiagram the merged ER diagram
	 */
	private void mergeGraphWithEntities(List<String> distinctMergedEntitySet, ERDiagram leftERDiagram, ERDiagram rightERDiagram, 
			ERDiagram connector, ERDiagram mergedERDiagram) {
		for(String entityName: distinctMergedEntitySet) {
			Entity entity = null;
			if(entityName.endsWith(LEFT_SUFFIX))
				entity = ERDiagramMergeUtil.lookupEntity(leftERDiagram, entityName.substring(0, entityName.length()-LEFT_SUFFIX.length()));
			else if(entityName.endsWith(RIGHT_SUFFIX)) 
				entity = ERDiagramMergeUtil.lookupEntity(rightERDiagram, entityName.substring(0, entityName.length()-RIGHT_SUFFIX.length()));
			else if(entityName.endsWith(CONNECTOR_SUFFIX))
				entity = ERDiagramMergeUtil.lookupEntity(connector, entityName.substring(0, entityName.length()-CONNECTOR_SUFFIX.length()));
			mergedERDiagram.getEntities().add(ERDiagramMergeUtil.cloneEntity(entity));
		}
	}
	
	
	/**
	 * Construct the final graph, add all merged relationships to the graph.
	 * @param distinctMergedRelationshipSet the list containing unique merged relationships
	 * @param mergedEntitySet the merged entity set
	 * @param distinctMergedEntitySet the list containing unique merged entities
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @param connector the connector ER diagram for the two input ER diagrams
	 * @param mergedERDiagram the merged ER diagram
	 */
	private void mergeGraphWithRelationships(List<String> distinctMergedRelationshipSet, List<List<String>> mergedEntitySet, 
			List<String> distinctMergedEntitySet, ERDiagram leftERDiagram, ERDiagram rightERDiagram, ERDiagram connector, ERDiagram mergedERDiagram) {		
		for(String relationshipName: distinctMergedRelationshipSet) {
			Relationship relationship = null;
			Relationship clonedRelationship = null;
			String fromEntityName = BLANK;
			String toEntityName = BLANK;
			String suffix = BLANK;

			if(relationshipName.endsWith(LEFT_SUFFIX)) {
				suffix = LEFT_SUFFIX;
				relationship = ERDiagramMergeUtil.lookupRelationship(leftERDiagram, 
						relationshipName.substring(0, relationshipName.length()-suffix.length()));
			}
			else if(relationshipName.endsWith(RIGHT_SUFFIX)) {
				suffix = RIGHT_SUFFIX;
				relationship = ERDiagramMergeUtil.lookupRelationship(rightERDiagram, 
						relationshipName.substring(0, relationshipName.length()-suffix.length()));
			}
			else if(relationshipName.endsWith(CONNECTOR_SUFFIX)) {
				suffix = CONNECTOR_SUFFIX;
				relationship = ERDiagramMergeUtil.lookupRelationship(connector, 
				relationshipName.substring(0, relationshipName.length()-suffix.length()));
			}
				
			clonedRelationship = ERDiagramMergeUtil.cloneRelationship(relationship);
				
			fromEntityName = ERDiagramMergeUtil.lookupMergedName(distinctMergedEntitySet, mergedEntitySet, relationship.getFrom().getName()+suffix);
			toEntityName = ERDiagramMergeUtil.lookupMergedName(distinctMergedEntitySet, mergedEntitySet, relationship.getTo().getName()+suffix); 
				
			clonedRelationship.setFrom(ERDiagramMergeUtil.lookupEntity(mergedERDiagram, fromEntityName.substring(0, fromEntityName.indexOf(AT))));
			clonedRelationship.setTo(ERDiagramMergeUtil.lookupEntity(mergedERDiagram, toEntityName.substring(0, toEntityName.indexOf(AT))));
				
			mergedERDiagram.getRelationships().add(clonedRelationship);
		}
	}
	
	
	/**
	 * Construct the final graph, add all merged entity attributes to the graph.
	 * @param distinctMergedEntityAttributeSet the list containing unique merged entity attributes
	 * @param mergedEntityAttributeSet merged entity attribute set
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @param connector the connector ER diagram for the two input ER diagrams
	 * @param mergedERDiagram the merged ER diagram
	 */
	private void mergeGraphWithEntityAttributes(List<String> distinctMergedEntityAttributeSet, List<List<String>> mergedEntityAttributeSet, 
			ERDiagram leftERDiagram, ERDiagram rightERDiagram, ERDiagram connector, ERDiagram mergedERDiagram) {	
		for(int round=1; round<3; round++) {
			for(String entityAttributeName: distinctMergedEntityAttributeSet) {		
				Attribute attribute = null;
				if(entityAttributeName.endsWith(LEFT_SUFFIX))
					attribute = ERDiagramMergeUtil.lookupEntityAttribute(leftERDiagram, 
							entityAttributeName.substring(0, entityAttributeName.length()-LEFT_SUFFIX.length()));
				else if(entityAttributeName.endsWith(RIGHT_SUFFIX))
					attribute = ERDiagramMergeUtil.lookupEntityAttribute(rightERDiagram, 
							entityAttributeName.substring(0, entityAttributeName.length()-RIGHT_SUFFIX.length()));
				else if(entityAttributeName.endsWith(CONNECTOR_SUFFIX))
					attribute = ERDiagramMergeUtil.lookupEntityAttribute(connector, 
							entityAttributeName.substring(0, entityAttributeName.length()-CONNECTOR_SUFFIX.length()));
				
				String mergedEntityAttributeName = ERDiagramMergeUtil.lookupMergedName(distinctMergedEntityAttributeSet, mergedEntityAttributeSet, 
						entityAttributeName);
				Entity entity = ERDiagramMergeUtil.lookupEntity(mergedERDiagram, mergedEntityAttributeName.substring(0, 
						mergedEntityAttributeName.indexOf(DASH)));
				
				if(round == 1 && !entityAttributeName.contains(SPLIT)) {
					entity.getEntityAttributes().add(ERDiagramMergeUtil.cloneAttribute(attribute));
				}
				else if(round == 2 && entityAttributeName.contains(SPLIT)){
					Attribute parentAttribute = ERDiagramMergeUtil.lookupEntityAttribute(mergedERDiagram, 
							entityAttributeName.substring(0, entityAttributeName.lastIndexOf(SPLIT)));
					parentAttribute.getAttributeAttributes().add(attribute);
				}
			}
		}
	}
	
	
	/**
	 * Construct the final graph, add all merged relationship attributes to the graph.
	 * @param distinctMergedRelationshipAttributeSet the list containing unique merged relationship attributes
	 * @param mergedRelationshipAttributeSet the merged relationship attributes
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @param connector the connector ER diagram for the two input ER diagrams
	 * @param mergedERDiagram the merged ER diagram
	 */
	private void mergeGraphWithRelationshipAttributes(List<String> distinctMergedRelationshipAttributeSet, List<List<String>> mergedRelationshipAttributeSet, 
			ERDiagram leftERDiagram, ERDiagram rightERDiagram, ERDiagram connector, ERDiagram mergedERDiagram) {
		for(String relationshipAttributeName: distinctMergedRelationshipAttributeSet) {
			Attribute attribute = null;
			if(relationshipAttributeName.endsWith(LEFT_SUFFIX))
				attribute = ERDiagramMergeUtil.lookupRelationshipAttribute(leftERDiagram, 
						relationshipAttributeName.substring(0, relationshipAttributeName.length()-LEFT_SUFFIX.length()));
			else if(relationshipAttributeName.endsWith(RIGHT_SUFFIX))
				attribute = ERDiagramMergeUtil.lookupRelationshipAttribute(rightERDiagram, 
						relationshipAttributeName.substring(0, relationshipAttributeName.length()-RIGHT_SUFFIX.length()));
			else if(relationshipAttributeName.endsWith(CONNECTOR_SUFFIX))
				attribute = ERDiagramMergeUtil.lookupRelationshipAttribute(connector, 
						relationshipAttributeName.substring(0, relationshipAttributeName.length()-CONNECTOR_SUFFIX.length()));
			
			String mergedRelationshipAttributeName = ERDiagramMergeUtil.lookupMergedName(distinctMergedRelationshipAttributeSet, 
					mergedRelationshipAttributeSet, relationshipAttributeName);
			Relationship relationship = ERDiagramMergeUtil.lookupRelationship(mergedERDiagram, 
					mergedRelationshipAttributeName.substring(0, mergedRelationshipAttributeName.lastIndexOf(DASH)));
			relationship.getRelationshipAttributes().add(ERDiagramMergeUtil.cloneAttribute(attribute));
		}
	}

	
	/**
	 * Validate whether the elements specified on two ends of the mapping really match or not. If yes, create the mapping hashMap, else throw exception. 
	 * @param modelRel the input relationship model specified by the end user
	 * @param leftERDiagram the left side input ER diagram
	 * @param rightERDiagram the right side input ER diagram
	 * @return mapping between the two input ER diagrams
	 * @throws Exception
	 */
	private Map<String, String> validateAndPopulateInputDiagramMapping(ModelRel modelRel, ERDiagram leftERDiagram, ERDiagram rightERDiagram) 
			throws Exception {
		Map<String, String> leftToRightDiagramMapping = new HashMap<String, String>();
		Map<String, String> leftUriToElemNameMapping = new HashMap<String, String>();
		Map<String, String> rightUriToElemNameMapping = new HashMap<String, String>();
		
		for(int i=0; i<modelRel.getLinks().size(); i++) {
			EList<ModelElementEndpoint> modelElementEndpoints = modelRel.getLinks().get(i).getModelElemEndpoints();
			if(!modelElementEndpoints.get(0).getTarget().getName().startsWith(ATTRIBUTE)) {
				leftUriToElemNameMapping.put(modelElementEndpoints.get(0).getTarget().getUri(), modelElementEndpoints.get(0).getTarget().getName());
				rightUriToElemNameMapping.put(modelElementEndpoints.get(1).getTarget().getUri(), modelElementEndpoints.get(1).getTarget().getName());	
			}
		}
		for(int i=0; i<modelRel.getLinks().size(); i++) {
			EList<ModelElementEndpoint> modelElementEndpoints = modelRel.getLinks().get(i).getModelElemEndpoints();
			String leftElemName = modelElementEndpoints.get(0).getTarget().getName();
			String leftElemUri = modelElementEndpoints.get(0).getTarget().getUri();
			String rightElemName = modelElementEndpoints.get(1).getTarget().getName();
			String rightElemUri = modelElementEndpoints.get(1).getTarget().getUri();
			if(leftElemName.startsWith(ENTITY)) {
				Entity leftEntity = leftERDiagram.getEntities().get(Integer.parseInt(leftElemUri.substring(leftElemUri.lastIndexOf(DOT)+1)));
				Entity rightEntity = rightERDiagram.getEntities().get(Integer.parseInt(rightElemUri.substring(rightElemUri.lastIndexOf(DOT)+1)));
				leftToRightDiagramMapping.put(ENTITY_PREFIX+leftElemName.substring(ENTITY.length()+1), 
						ENTITY_PREFIX+rightElemName.substring(ENTITY.length()+1));
				if (!isOpenWorld)
					ERDiagramMergeUtil.validateEquality(leftEntity, rightEntity);
			}
			else if(leftElemName.startsWith(RELATIONSHIP)) {
				Relationship leftRelationship = leftERDiagram.getRelationships().get(Integer.parseInt(leftElemUri.substring(leftElemUri.lastIndexOf(DOT)+1)));
				Relationship rightRelationship = rightERDiagram.getRelationships()
						.get(Integer.parseInt(rightElemUri.substring(rightElemUri.lastIndexOf(DOT)+1)));
				leftToRightDiagramMapping.put(RELATIONSHIP_PREFIX+leftRelationship.getFrom().getName()+DASH
						+leftElemName.substring(RELATIONSHIP.length()+1)+DASH
						+leftRelationship.getTo().getName(), RELATIONSHIP_PREFIX+rightRelationship.getFrom().getName()+DASH
						+rightElemName.substring(RELATIONSHIP.length()+1)+DASH+rightRelationship.getTo().getName());
				if (!isOpenWorld)
					ERDiagramMergeUtil.validateEquality(leftRelationship, rightRelationship);
			}
			else if(leftElemName.startsWith(ATTRIBUTE)) {
				String leftEntityOrRelationshipUri = leftElemUri.substring(0, leftElemUri.lastIndexOf(AT)-1);
				String leftEntityOrRelationshipName = leftUriToElemNameMapping.get(leftEntityOrRelationshipUri);
				String rightEntityOrRelationshipUri = rightElemUri.substring(0, rightElemUri.lastIndexOf(AT)-1);
				String rightEntityOrRelationshipName = rightUriToElemNameMapping.get(rightEntityOrRelationshipUri);
				
				if(leftEntityOrRelationshipName.startsWith(ENTITY)) {
					Entity leftAssociatedEntity = leftERDiagram.getEntities()
							.get(Integer.parseInt(leftEntityOrRelationshipUri.substring(leftEntityOrRelationshipUri.lastIndexOf(DOT)+1)));
					Entity rightAssociatedEntity = rightERDiagram.getEntities()
							.get(Integer.parseInt(rightEntityOrRelationshipUri.substring(rightEntityOrRelationshipUri.lastIndexOf(DOT)+1)));
					String leftAttributeName = leftAssociatedEntity.getName()+DASH+leftElemName.substring(ATTRIBUTE.length()+1);
					String rightAttributeName = rightAssociatedEntity.getName()+DASH+rightElemName.substring(ATTRIBUTE.length()+1);
					leftToRightDiagramMapping.put(ENTITY_ATTRIBUTE_PREFIX+leftAttributeName, ENTITY_ATTRIBUTE_PREFIX+rightAttributeName);
					
					Attribute leftAttribute = ERDiagramMergeUtil.lookupEntityAttribute(leftERDiagram, leftAttributeName);
					Attribute rightAttribute = ERDiagramMergeUtil.lookupEntityAttribute(rightERDiagram, rightAttributeName);
					if (!isOpenWorld) {
						ERDiagramMergeUtil.validateEquality(leftAssociatedEntity, rightAssociatedEntity);
						ERDiagramMergeUtil.validateEquality(leftAssociatedEntity, leftAttribute, rightAssociatedEntity, rightAttribute);	
					}
				}
				else if(leftEntityOrRelationshipName.startsWith(RELATIONSHIP)) {
					Relationship leftAssociatedRelationship = leftERDiagram.getRelationships()
							.get(Integer.parseInt(leftEntityOrRelationshipUri.substring(leftEntityOrRelationshipUri.lastIndexOf(DOT)+1)));
					Relationship rightAssociatedRelationship = rightERDiagram.getRelationships()
							.get(Integer.parseInt(rightEntityOrRelationshipUri.substring(rightEntityOrRelationshipUri.lastIndexOf(DOT)+1)));
					
					String leftAttributeName = leftAssociatedRelationship.getFrom().getName()+DASH
							+leftEntityOrRelationshipName.substring(RELATIONSHIP.length()+1)+DASH+leftAssociatedRelationship.getTo().getName()+DASH
							+leftElemName.substring(ATTRIBUTE.length()+1);
					String rightAttributeName = rightAssociatedRelationship.getFrom().getName()+DASH
							+rightEntityOrRelationshipName.substring(RELATIONSHIP.length()+1)+DASH+rightAssociatedRelationship.getTo().getName()+DASH
							+rightElemName.substring(ATTRIBUTE.length()+1);
					leftToRightDiagramMapping.put(RELATIONSHIP_ATTRIBUTE_PREFIX+leftAttributeName, RELATIONSHIP_ATTRIBUTE_PREFIX+rightAttributeName);
					
					Attribute leftAttribute = ERDiagramMergeUtil.lookupRelationshipAttribute(leftERDiagram, leftAttributeName);
					Attribute rightAttribute = ERDiagramMergeUtil.lookupRelationshipAttribute(rightERDiagram, rightAttributeName);
					if (!isOpenWorld) {
						ERDiagramMergeUtil.validateEquality(leftAssociatedRelationship, rightAssociatedRelationship);
						ERDiagramMergeUtil.validateEquality(leftAssociatedRelationship, leftAttribute, rightAssociatedRelationship, rightAttribute);	
					}
				}
			}
		}
		return leftToRightDiagramMapping;
	}
	
	
	/**
	 * Read the properties defined in the .properties file.
	 * @param properties 
	 * @throws Exception
	 */
	private void readProperties(Properties properties) throws Exception {
		isOpenWorld = MultiModelOperatorUtils.getBoolProperty(properties, PROPERTY_IN_ISOPENWORLD);
		preferRightModelNaming = MultiModelOperatorUtils.getBoolProperty(properties, PROPERTY_IN_PREFERRIGHTMODELNAMING); 
	}

}
