/*
 * Copyright (c) 2012 Marsha Chechik, Alessio Di Sandro, Michalis Famelis,
 * Rick Salay, Vivien Suen.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    Wentao He - Implementation.
 */
package edu.toronto.cs.se.modelepedia.erdiagram.operator.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edu.toronto.cs.se.modelepedia.erdiagram.Attribute;
import edu.toronto.cs.se.modelepedia.erdiagram.ERDiagram;
import edu.toronto.cs.se.modelepedia.erdiagram.Entity;
import edu.toronto.cs.se.modelepedia.erdiagram.ErdiagramFactory;
import edu.toronto.cs.se.modelepedia.erdiagram.Relationship;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.enumeration.LevelEnum;

/**
 * Utility methods for the ER diagram merge.
 */
public class ERDiagramMergeUtil {
	
	private static final String DASH = "-";
	private static final String ENTITY_PREFIX = "ENTITY@";
	private static final String RELATIONSHIP_PREFIX = "RELATIONSHIP@";
	private static final String ENTITY_ATTRIBUTE_PREFIX = "ENTITY_ATTRIBUTE@";
	private static final String RELATIONSHIP_ATTRIBUTE_PREFIX = "RELATIONSHIP_ATTRIBUTE@";
	
	/**
	 * Lookup an entity from an ER diagram by entity name. 
	 * Entity name format should be: entityName
	 * @param erdiagram
	 * @param name
	 * @return the looked up entity
	 */
	public static Entity lookupEntity(ERDiagram erdiagram, String name) {
		Entity entity = null;
		for(Entity e: erdiagram.getEntities()) {
			if(name.equals(e.getName())) {
				entity = e;
				break;
			}
		}
		return entity;
	}	
	
	
	/**
	 * Lookup a relationship from an ER diagram by relationship name.
	 * Relationship name format should be: fromEntityName-relationshipName-toEntityName
	 * @param erdiagram
	 * @param name
	 * @return the looked up relationship
	 */
	public static Relationship lookupRelationship(ERDiagram erdiagram, String name) {
		Relationship relationship = null;
		String toEntityName = name.substring(name.lastIndexOf(DASH)+1);
		name = name.substring(0, name.lastIndexOf(DASH));
		String relationshipName = name.substring(name.lastIndexOf(DASH)+1);
		String fromEntityName = name.substring(0, name.lastIndexOf(DASH));
		
		for(Relationship r: erdiagram.getRelationships()) {
			if(relationshipName.equals(r.getName()) && fromEntityName.equals(r.getFrom().getName()) && toEntityName.equals(r.getTo().getName())) {
				relationship = r;
				break;
			}
		}
		return relationship;
	}
	
	
	/**
	 * Lookup an entity attribute from an ER diagram by attribute name.
	 * Attribute name format should be: entityName-attributeName, entityName-attributeName^subAttributeName
	 * @param erdiagram
	 * @param name
	 * @return the looked up attribute
	 */
	public static Attribute lookupEntityAttribute(ERDiagram erdiagram, String name) {
		Attribute attribute = null;
		String entityName = name.substring(0, name.lastIndexOf(DASH));
		String attributeName = name.substring(name.lastIndexOf(DASH)+1);
		String subAttributeName = "";
		
		for(Entity entity: erdiagram.getEntities()) {
			if(entityName.equals(entity.getName())) {
				for(Attribute a: entity.getEntityAttributes()) {
					if(attributeName.contains("^") && a.getName().equals(attributeName.substring(0, attributeName.lastIndexOf("^")))) {
						subAttributeName = attributeName.substring(attributeName.lastIndexOf("^")+1);
						attributeName = attributeName.substring(0, attributeName.lastIndexOf("^"));
						for(Attribute subAttribute: a.getAttributeAttributes()) {
							if(subAttributeName.equals(subAttribute.getName())) {
								attribute = subAttribute;
								break;
							}
						}
					}
					else if(attributeName.equals(a.getName())) {
						attribute = a;
					}
				}
			}
		}
		return attribute;
	}
	
	
	/**
	 * Lookup a relationship attribute from an ER diagram by attribute name.
	 * Attribute name format should be: fromEntityName-relationshipName-toEntityName-attributeName
	 * @param erdiagram
	 * @param name
	 * @return the looked up attribute
	 */
	public static Attribute lookupRelationshipAttribute(ERDiagram erdiagram, String name) {
		Attribute attribute = null;
		String relationshipName = name.substring(0, name.lastIndexOf(DASH));
		String attributeName = name.substring(name.lastIndexOf(DASH)+1);
		
		for(Relationship relationship: erdiagram.getRelationships()) {
			if(relationshipName.equals(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName())) {
				for(Attribute a: relationship.getRelationshipAttributes()) {
					if(attributeName.equals(a.getName())) {
						attribute = a;
					}
				}
			}
		}
		return attribute;
	}
	
	
	/**
	 * Lookup the element in the merged set who is the final merged element.
	 * @param distinctMergedSet
	 * @param mergedSet
	 * @param name
	 * @return the element who is the final merged element
	 */
	public static String lookupMergedName(List<String> distinctMergedSet, List<List<String>> mergedSet, String name) {
		List<String> list = new ArrayList<String>();
		for(List<String> li: mergedSet) {
			for(String item: li) {
				if(name.equals(item)) {
					list = li;
				}
			}
		}
		for(String item: list) {
			if(distinctMergedSet.contains(item)) {
				return item;
			}
		}
		return null;
	}
	
	
	/**
	 * Clone an entity.
	 * Note: Did not put this in Entity class because it is auto-generated. Avoid making changes to auto-generated classes.
	 * @param entity
	 * @return a same entity as the input entity
	 */
	public static Entity cloneEntity(Entity entity) {
		Entity clonedEntity = ErdiagramFactory.eINSTANCE.createEntity();
		clonedEntity.setName(entity.getName());
		clonedEntity.setWeak(entity.isWeak());
		return clonedEntity;
	}
	
	
	/**
	 * Clone a relationship.
	 * Note: Did not put this in Relationship class because it is auto-generated. Avoid making changes to auto-generated classes.
	 * @param relationship
	 * @return a same relationship as the input relationship
	 */
	public static Relationship cloneRelationship(Relationship relationship) {
		Relationship clonedRelationship = ErdiagramFactory.eINSTANCE.createRelationship();
		clonedRelationship.setName(relationship.getName());
		clonedRelationship.setCardinality(relationship.getCardinality());
		clonedRelationship.setWeak(relationship.isWeak());
		return clonedRelationship;
	}
	
	
	/**
	 * Clone an attribute. 
	 * Note: Did not put this in Attribute class because it is auto-generated. Avoid making changes to auto-generated classes.
	 * @param attribute
	 * @return a same attribute as the input attribute
	 */
	public static Attribute cloneAttribute(Attribute attribute) {
		Attribute clonedAttribute = ErdiagramFactory.eINSTANCE.createAttribute();
		clonedAttribute.setName(attribute.getName());
		clonedAttribute.setMultivalued(attribute.isMultivalued());
		clonedAttribute.setKey(attribute.isKey());
		clonedAttribute.setDerived(attribute.isDerived());
		clonedAttribute.setComposite(attribute.isComposite());
		return clonedAttribute;
	}
	
	
	/**
	 * Validate if two entities are the same or not. If not, throw exception.
	 * @param entity1
	 * @param entity2
	 * @throws Exception
	 */
	public static void validateEquality(Entity entity1, Entity entity2) throws Exception {
		if(entity1.isWeak() != entity2.isWeak()) {
			throw new Exception("Left entity '" + entity1.getName() + "' and right entity '" + entity2.getName() + "' have different properties." + "\n");
		}
	}
	
	
	/**
	 * Validate if two relationships are the same or not. If not, throw exception.
	 * @param relationship1
	 * @param relationship2
	 * @throws Exception
	 */
	public static void validateEquality(Relationship relationship1, Relationship relationship2) throws Exception {
		if(relationship1.isWeak() != relationship2.isWeak()
				|| !relationship1.getCardinality().equals(relationship2.getCardinality())) {
			throw new Exception("Left relationship '" + relationship1.getName() + "' and right relationship '" + relationship2.getName() + 
					"' have different properties." + "\n");
		}
	}	
	
	
	/**
	 * Validate if two attributes are the same or not. If not, throw exception.
	 * @param entity1 the entity that attribute1 is associated to
	 * @param attribute1
	 * @param entity2 the entity that attribute2 is associated to
	 * @param attribute2
	 * @throws Exception
	 */
	public static void validateEquality(Entity entity1, Attribute attribute1, Entity entity2, Attribute attribute2) throws Exception {
		if(attribute1.isComposite() != attribute2.isComposite()
				|| attribute1.isDerived() != attribute2.isDerived()
				|| attribute1.isKey() != attribute2.isKey()
				|| attribute1.isMultivalued() != attribute2.isMultivalued()) {
			throw new Exception("Left attribute '" + attribute1.getName() + "' of entity '" + entity1.getName() + 
					"' and right attribute '" + attribute2.getName() + "' of entity '" + entity2.getName() + "' have different properties." + "\n");
		}
	}
	
	
	/**
	 * Validate if two attributes are the same or not. If not, throw exception.
	 * @param relationship1 the relationship that attribute1 is associated to
	 * @param attribute1
	 * @param relationship2 the relationship that attribute2 is associated to
	 * @param attribute2
	 * @throws Exception
	 */
	public static void validateEquality(Relationship relationship1, Attribute attribute1, Relationship relationship2, Attribute attribute2) throws Exception {
		if(attribute1.isComposite() != attribute2.isComposite()
				|| attribute1.isDerived() != attribute2.isDerived()
				|| attribute1.isKey() != attribute2.isKey()
				|| attribute1.isMultivalued() != attribute2.isMultivalued()) {
			throw new Exception("Left attribute '" + attribute1.getName() + "' of relationship '" + relationship1.getName() + 
					"' and right attribute '" + attribute2.getName() + "' of relationship '" + relationship2.getName() 
					+ "' have different properties." + "\n");
		}
	}
	
	
	/**
	 * Retrieve entity level mapping.
	 * @param diagramMapping
	 * @return entity level mapping
	 */
	public static Map<String, String> getEntityMapping(Map<String, String> diagramMapping) {
		return getMapping(diagramMapping, LevelEnum.ENTITY);
	}
	
	
	/**
	 * Retrieve relationship level mapping.
	 * @param diagramMapping
	 * @return relationship level mapping
	 */
	public static Map<String, String> getRelationshipMapping(Map<String, String> diagramMapping) {
		return getMapping(diagramMapping, LevelEnum.RELTIONSHIP);
	}
	
	
	/**
	 * Retrieve entity attribute level mapping.
	 * @param diagramMapping
	 * @return entity attribute level mapping
	 */
	public static Map<String, String> getEntityAttributeMapping(Map<String, String> diagramMapping) {
		return getMapping(diagramMapping, LevelEnum.ATTRIBUTE_OF_ENTITY);
	}
	

	/**
	 * Retrieve relationship attribute level mapping.
	 * @param diagramMapping
	 * @return relationship attribute level mapping
	 */
	public static Map<String, String> getRelationshipAttributeMapping(Map<String, String> diagramMapping) {
		return getMapping(diagramMapping, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP);
	}
	
	
	/**
	 * Retrieve mapping of a specified element level.
	 * @param diagramMapping
	 * @param level
	 * @return mapping of the specified element level
	 */
	private static Map<String, String> getMapping(Map<String, String> diagramMapping, LevelEnum level) {
		Map<String, String> mapping = new HashMap<String, String>();
		for(int i=0; i<diagramMapping.size(); i++) {
			String key = (String)diagramMapping.keySet().toArray()[i];
			if((LevelEnum.ENTITY.equals(level) && key.startsWith(ENTITY_PREFIX))
					|| (LevelEnum.RELTIONSHIP.equals(level) && key.startsWith(RELATIONSHIP_PREFIX))
					|| (LevelEnum.ATTRIBUTE_OF_ENTITY.equals(level) && key.startsWith(ENTITY_ATTRIBUTE_PREFIX))
					|| (LevelEnum.ATTRIBUTE_OF_RELATIONSHIP.equals(level) && key.startsWith(RELATIONSHIP_ATTRIBUTE_PREFIX))) {
					mapping.put(key, diagramMapping.get(key));
			}
		}	
		return mapping;		
	}
	
	
	/**
	 * For a list, convert each item in the list to a list containing only one item, so that form a list of list. 
	 * @param list
	 * @return List<List<String>>
	 */
	public static List<List<String>> convertToListOfList(List<String> list) {
		List<List<String>> listOfList = new ArrayList<List<String>>();
		for(String item: list) {
			List<String> itemWrapper = new ArrayList<String>();
			itemWrapper.add(item);
			listOfList.add(itemWrapper);
		}
		return listOfList;
	}
	
	
	/**
	 * Merge two lists in a list of list. The list containing "to" value will be removed, the value will be added to the list containing "from" value.
	 * @param listOfList
	 * @param from
	 * @param to
	 */
	public static void mergeList(List<List<String>> listOfList, String from, String to) {
		List<String> listContainingFromItem = null; 
		List<String> listContainingToItem = null;
		for(List<String> list: listOfList) {
			if(list.contains(from)) {
				listContainingFromItem = list;
			}
			else if(list.contains(to)) {
				listContainingToItem = list;
			}
		}
		
		listOfList.remove(listContainingToItem);
		listContainingFromItem.addAll(listContainingToItem);
	}

	
}
