/*
 * Copyright (c) 2012 Marsha Chechik, Alessio Di Sandro, Michalis Famelis,
 * Rick Salay, Vivien Suen.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    Wentao He - Implementation.
 */
package edu.toronto.cs.se.modelepedia.erdiagram.operator;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import edu.toronto.cs.se.modelepedia.erdiagram.Attribute;
import edu.toronto.cs.se.modelepedia.erdiagram.ERDiagram;
import edu.toronto.cs.se.modelepedia.erdiagram.Entity;
import edu.toronto.cs.se.modelepedia.erdiagram.ErdiagramFactory;
import edu.toronto.cs.se.modelepedia.erdiagram.Relationship;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.enumeration.LevelEnum;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.enumeration.SourceEnum;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.util.ERDiagramMergeUtil;

import junit.framework.TestCase;

/**
 * Junit test cases for methods in ERDiagramMerge.java
 * Use java reflection to access private methods in the target class.
 */
public class TestERDiagramMerge extends TestCase {
			
	private static final String METHOD_CREATE_CONNECTOR = "createConnector";
	private static final String METHOD_CREATE_CONNECTOR_TO_DIAGRAM_MAPPING = "createConnectorToDiagramMapping";
	private static final String METHOD_POPULATE_DISJOINT_UNION = "populateDisjointUnion";
	private static final String METHOD_MERGE_SET = "mergeSet";
	private static final String METHOD_YIELD_DISTINCT_MERGED_SET = "yieldDistinctMergedSet";
	private static final String METHOD_MERGE_GRAPH_WITH_ENTITIES = "mergeGraphWithEntities";
	private static final String METHOD_MERGE_GRAPH_WITH_RELATIONSHIPS = "mergeGraphWithRelationships";
	private static final String METHOD_MERGE_GRAPH_WITH_ENTITY_ATTRIBUTES = "mergeGraphWithEntityAttributes";
	private static final String METHOD_MERGE_GRAPH_WITH_RELATIONSHIP_ATTRIBUTES = "mergeGraphWithRelationshipAttributes";
	private static final String METHOD_MERGE = "merge";
	
	private static final String DASH = "-";
	
	ERDiagramMerge erdiagramMerge;
	ErdiagramFactory erdiagramFactory;
	
	ERDiagram leftERDiagram;
	ERDiagram rightERDiagram;
	ERDiagram connector;
	ERDiagram mergedERDiagram;
	
	Map<String, String> leftToRightDiagramMapping; 
	Map<String, String> connectorToLeftDiagramEntityMapping;
	Map<String, String> connectorToLeftDiagramRelationshipMapping;
	Map<String, String> connectorToLeftDiagramEntityAttributeMapping;
	Map<String, String> connectorToLeftDiagramRelationshipAttributeMapping;
	Map<String, String> connectorToRightDiagramEntityMapping;
	Map<String, String> connectorToRightDiagramRelationshipMapping;
	Map<String, String> connectorToRightDiagramEntityAttributeMapping;
	Map<String, String> connectorToRightDiagramRelationshipAttributeMapping;
	
	List<String> entityDisjointUnion;
	List<String> relationshipDisjointUnion;
	List<String> entityAttributeDisjointUnion;
	List<String> relationshipAttributeDisjointUnion;
	
	protected void setUp() throws Exception {		
		erdiagramMerge = new ERDiagramMerge();
		erdiagramFactory = ErdiagramFactory.eINSTANCE;
		
		leftERDiagram = ERDiagramTestDataGenerator.prepareLeftERDiagram();
		rightERDiagram = ERDiagramTestDataGenerator.prepareRightERDiagram();
		connector = ERDiagramTestDataGenerator.prepareConnector();
		mergedERDiagram = erdiagramFactory.createERDiagram();
		
		leftToRightDiagramMapping = ERDiagramTestDataGenerator.prepareLeftToRightDiagramMapping();
		connectorToLeftDiagramEntityMapping = ERDiagramTestDataGenerator.prepareConnectorToLeftDiagramEntityMapping();
		connectorToLeftDiagramRelationshipMapping = ERDiagramTestDataGenerator.prepareConnectorToLeftDiagramRelationshipMapping();
		connectorToLeftDiagramEntityAttributeMapping = ERDiagramTestDataGenerator.prepareConnectorToLeftDiagramEntityAttributeMapping();
		connectorToLeftDiagramRelationshipAttributeMapping = ERDiagramTestDataGenerator.prepareConnectorToLeftDiagramRelationshipAttributeMapping();
		connectorToRightDiagramEntityMapping = ERDiagramTestDataGenerator.prepareConnectorToRightDiagramEntityMapping();
		connectorToRightDiagramRelationshipMapping = ERDiagramTestDataGenerator.prepareConnectorToRightDiagramRelationshipMapping();
		connectorToRightDiagramEntityAttributeMapping = ERDiagramTestDataGenerator.prepareConnectorToRightDiagramEntityAttributeMapping();
		connectorToRightDiagramRelationshipAttributeMapping = ERDiagramTestDataGenerator.prepareConnectorToRightDiagramRelationshipAttributeMapping();
		
		entityDisjointUnion = ERDiagramTestDataGenerator.prepareEntityDisjointUnion();
		relationshipDisjointUnion = ERDiagramTestDataGenerator.prepareRelationshipDisjointUnion();
		entityAttributeDisjointUnion = ERDiagramTestDataGenerator.prepareEntityAttributeDisjointUnion();
		relationshipAttributeDisjointUnion = ERDiagramTestDataGenerator.prepareRelationshipAttributeDisjointUnion();
	}
	
	
	public void testCreateConnector() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_CREATE_CONNECTOR, new Class[]{ERDiagram.class, Map.class});
			method.setAccessible(true); 
			ERDiagram connector = (ERDiagram)method.invoke(erdiagramMerge, new Object[]{rightERDiagram, leftToRightDiagramMapping});
			method.setAccessible(false);
			
			List<String> entityNameList = new ArrayList<String>();
			for(Entity entity: connector.getEntities()) {
				entityNameList.add(entity.getName());
			}
			
			assertEquals(3, entityNameList.size());
			assertTrue(entityNameList.contains("Employee"));
			assertTrue(entityNameList.contains("Department"));
			assertTrue(entityNameList.contains("Project"));
			
			List<Attribute> attributesOfEmployee = ERDiagramMergeUtil.lookupEntity(connector, "Employee").getEntityAttributes();
			List<String> attributeNamesOfEmployee = new ArrayList<String>();
			for(Attribute attribute: attributesOfEmployee) 
				attributeNamesOfEmployee.add(attribute.getName());
			
			List<Attribute> attributesOfDepartment = ERDiagramMergeUtil.lookupEntity(connector, "Department").getEntityAttributes();
			List<String> attributeNamesOfDepartment = new ArrayList<String>();
			for(Attribute attribute: attributesOfDepartment) 
				attributeNamesOfDepartment.add(attribute.getName());
			
			List<Attribute> attributesOfProject = ERDiagramMergeUtil.lookupEntity(connector, "Project").getEntityAttributes();
			List<String> attributeNamesOfProject = new ArrayList<String>();
			for(Attribute attribute: attributesOfProject)
				attributeNamesOfProject.add(attribute.getName());
			
			assertEquals(1, attributesOfEmployee.size());
			assertTrue(attributeNamesOfEmployee.contains("emp_id"));
			
			assertEquals(2, attributesOfDepartment.size());
			assertTrue(attributeNamesOfDepartment.contains("name"));
			assertTrue(attributeNamesOfDepartment.contains("employee_count"));
			
			assertEquals(1, attributesOfProject.size());
			assertTrue(attributeNamesOfProject.contains("name"));
			
			List<String> relationshipNameList = new ArrayList<String>();
			for(Relationship relationship: connector.getRelationships()) {
				relationshipNameList.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName());
			}
			
			assertEquals(2, relationshipNameList.size());
			assertTrue(relationshipNameList.contains("Employee-works_for-Department"));
			assertTrue(relationshipNameList.contains("Department-owns-Project"));
			
			List<Attribute> attributesOfWorksFor = ERDiagramMergeUtil.lookupRelationship(connector, 
					"Employee-works_for-Department").getRelationshipAttributes();
			List<String> attributeNamesOfWorksFor = new ArrayList<String>();
			for(Attribute attribute: attributesOfWorksFor)
				attributeNamesOfWorksFor.add(attribute.getName());
			
			assertEquals(1, attributeNamesOfWorksFor.size());
			assertTrue(attributeNamesOfWorksFor.contains("since"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }	
	}
	
	
	public void testCreateConnectorToDiagramMapping_WithRightDiagram() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_CREATE_CONNECTOR_TO_DIAGRAM_MAPPING, new Class[]{Map.class, SourceEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			Map<String, String> connectorToRightDiagramMapping = (Map<String, String>)method.invoke(erdiagramMerge, 
					new Object[]{leftToRightDiagramMapping, SourceEnum.RIGHT_DIAGRAM});
			method.setAccessible(false);
			
			assertEquals(10, connectorToRightDiagramMapping.size());
			
			for(int i=0; i<connectorToRightDiagramMapping.size(); i++) {
				String key = (String)connectorToRightDiagramMapping.keySet().toArray()[i];
				assertEquals(key, connectorToRightDiagramMapping.get(key));
			}
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }	
	}
	
	
	public void testCreateConnectorToDiagramMapping_WithLeftDiagram() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_CREATE_CONNECTOR_TO_DIAGRAM_MAPPING, new Class[]{Map.class, SourceEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			Map<String, String> connectorToLeftDiagramMapping = (Map<String, String>)method.invoke(erdiagramMerge, 
					new Object[]{leftToRightDiagramMapping, SourceEnum.LEFT_DIAGRAM});
			method.setAccessible(false);
			
			assertEquals(10, connectorToLeftDiagramMapping.size());
			
			for(int i=0; i<leftToRightDiagramMapping.size(); i++) {
				String key = (String)leftToRightDiagramMapping.keySet().toArray()[i];
				String value = (String)leftToRightDiagramMapping.values().toArray()[i];
				assertEquals(connectorToLeftDiagramMapping.get(value), key);
			}
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }	
	}
	
	
	public void testPopulateDisjointUnion_Entity() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_POPULATE_DISJOINT_UNION, 
					new Class[]{ERDiagram.class, ERDiagram.class, ERDiagram.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> disjointUnion = (List<String>)method.invoke(erdiagramMerge, 
					new Object[]{leftERDiagram, rightERDiagram, connector, LevelEnum.ENTITY});
			method.setAccessible(false);
			
			assertEquals(11, disjointUnion.size());
			assertTrue(disjointUnion.contains("Employee@Left"));
			assertTrue(disjointUnion.contains("Unit@Left"));
			assertTrue(disjointUnion.contains("Project@Left"));
			assertTrue(disjointUnion.contains("Dependent@Left"));
			assertTrue(disjointUnion.contains("Employee@Right"));
			assertTrue(disjointUnion.contains("Department@Right"));
			assertTrue(disjointUnion.contains("Project@Right"));
			assertTrue(disjointUnion.contains("Module@Right"));
			assertTrue(disjointUnion.contains("Employee@Connector"));
			assertTrue(disjointUnion.contains("Department@Connector"));
			assertTrue(disjointUnion.contains("Project@Connector"));
		
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }			
	}
	
	
	public void testPopulateDisjointUnion_Relationship() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_POPULATE_DISJOINT_UNION, 
					new Class[]{ERDiagram.class, ERDiagram.class, ERDiagram.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> disjointUnion = (List<String>)method.invoke(erdiagramMerge, 
					new Object[]{leftERDiagram, rightERDiagram, connector, LevelEnum.RELTIONSHIP});
			method.setAccessible(false);
			
			assertEquals(11, disjointUnion.size());
			assertTrue(disjointUnion.contains("Employee-belongs_to-Unit@Left"));
			assertTrue(disjointUnion.contains("Unit-has-Project@Left"));
			assertTrue(disjointUnion.contains("Employee-works_on-Project@Left"));
			assertTrue(disjointUnion.contains("Employee-dependent_of-Dependent@Left"));
			assertTrue(disjointUnion.contains("Employee-manages-Department@Right"));
			assertTrue(disjointUnion.contains("Employee-works_for-Department@Right"));
			assertTrue(disjointUnion.contains("Department-owns-Project@Right"));
			assertTrue(disjointUnion.contains("Project-contains-Module@Right"));
			assertTrue(disjointUnion.contains("Employee-supervises-Employee@Right"));
			assertTrue(disjointUnion.contains("Employee-works_for-Department@Connector"));
			assertTrue(disjointUnion.contains("Department-owns-Project@Connector"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }			
	}
	
	
	public void testPopulateDisjointUnion_EntityAttribute() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_POPULATE_DISJOINT_UNION, 
					new Class[]{ERDiagram.class, ERDiagram.class, ERDiagram.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> disjointUnion = (List<String>)method.invoke(erdiagramMerge, 
					new Object[]{leftERDiagram, rightERDiagram, connector, LevelEnum.ATTRIBUTE_OF_ENTITY});
			method.setAccessible(false);
			
			assertEquals(25, disjointUnion.size());
			assertTrue(disjointUnion.contains("Employee-id@Left"));
			assertTrue(disjointUnion.contains("Employee-date_of_birth@Left"));
			assertTrue(disjointUnion.contains("Employee-age@Left"));
			assertTrue(disjointUnion.contains("Employee-name@Left"));
			assertTrue(disjointUnion.contains("Employee-name^last_name@Left"));
			assertTrue(disjointUnion.contains("Employee-name^first_name@Left"));
			assertTrue(disjointUnion.contains("Unit-head_count@Left"));
			assertTrue(disjointUnion.contains("Unit-name@Left"));
			assertTrue(disjointUnion.contains("Project-title@Left"));
			assertTrue(disjointUnion.contains("Dependent-name@Left"));
			assertTrue(disjointUnion.contains("Dependent-date_of_birth@Left"));
			assertTrue(disjointUnion.contains("Dependent-relationship@Left"));
			assertTrue(disjointUnion.contains("Employee-emp_id@Right"));
			assertTrue(disjointUnion.contains("Department-employee_count@Right"));
			assertTrue(disjointUnion.contains("Department-name@Right"));
			assertTrue(disjointUnion.contains("Department-locations@Right"));
			assertTrue(disjointUnion.contains("Project-size@Right"));
			assertTrue(disjointUnion.contains("Project-name@Right"));
			assertTrue(disjointUnion.contains("Project-start_date@Right"));
			assertTrue(disjointUnion.contains("Module-name@Right"));
			assertTrue(disjointUnion.contains("Module-start_date@Right"));
			assertTrue(disjointUnion.contains("Employee-emp_id@Connector"));
			assertTrue(disjointUnion.contains("Department-employee_count@Connector"));
			assertTrue(disjointUnion.contains("Department-name@Connector"));
			assertTrue(disjointUnion.contains("Project-name@Connector"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }			
	}
	
	
	public void testPopulateDisjointUnion_RelationshipAttribute() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_POPULATE_DISJOINT_UNION, 
					new Class[]{ERDiagram.class, ERDiagram.class, ERDiagram.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> disjointUnion = (List<String>)method.invoke(erdiagramMerge, 
					new Object[]{leftERDiagram, rightERDiagram, connector, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP});
			method.setAccessible(false);
			
			assertEquals(4, disjointUnion.size());
			assertTrue(disjointUnion.contains("Employee-belongs_to-Unit-start_date@Left"));
			assertTrue(disjointUnion.contains("Employee-manages-Department-start_date@Right"));
			assertTrue(disjointUnion.contains("Employee-works_for-Department-since@Right"));
			assertTrue(disjointUnion.contains("Employee-works_for-Department-since@Connector"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }			
	}
	
	
	public void testMergeSet_Entity() {
		try {			
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntitySet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityMapping, connectorToRightDiagramEntityMapping, entityDisjointUnion, LevelEnum.ENTITY});
			method.setAccessible(false);
			
			assertEquals(5, mergedEntitySet.size());
			for(int i=0; i<mergedEntitySet.size(); i++) {
				List<String> list = mergedEntitySet.get(i);
				if(list.contains("Employee@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Employee@Connector"));
					assertTrue(list.contains("Employee@Right"));
				}
				else if(list.contains("Unit@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Department@Connector"));
					assertTrue(list.contains("Department@Right"));
				}
				else if(list.contains("Project@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Project@Connector"));
					assertTrue(list.contains("Project@Right"));
				}
				else if(list.contains("Dependent@Left")) {
					assertEquals(1, list.size());
				}
				else if(list.contains("Module@Right")) {
					assertEquals(1, list.size());
				}
				else {
					assertTrue(false); //Ideally it should never get here. If it gets here, means there is bug! 
				}
			}
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testMergeSet_Relationship() {
		try {			
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedRelationshipSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramRelationshipMapping, connectorToRightDiagramRelationshipMapping, 
					relationshipDisjointUnion, LevelEnum.RELTIONSHIP});
			method.setAccessible(false);
			
			assertEquals(7, mergedRelationshipSet.size());
			for(int i=0; i<mergedRelationshipSet.size(); i++) {
				List<String> list = mergedRelationshipSet.get(i);
				if(list.contains("Employee-belongs_to-Unit@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Employee-works_for-Department@Connector"));
					assertTrue(list.contains("Employee-works_for-Department@Right"));
				}
				else if(list.contains("Unit-has-Project@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Department-owns-Project@Connector"));
					assertTrue(list.contains("Department-owns-Project@Right"));
				}
				else if(list.contains("Employee-works_on-Project@Left") 
						|| list.contains("Employee-dependent_of-Dependent@Left")
						|| list.contains("Employee-manages-Department@Right")
						|| list.contains("Project-contains-Module@Right")
						|| list.contains("Employee-supervises-Employee@Right")) {
					assertEquals(1, list.size());
				}
				else {
					assertTrue(false); //Ideally it should never get here. If it gets here, means there is bug! 
				}
			}
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	
	public void testMergeSet_EntityAttribute() {
		try {			
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntityAttributeSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityAttributeMapping, connectorToRightDiagramEntityAttributeMapping, 
					entityAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_ENTITY});
			method.setAccessible(false);

			assertEquals(17, mergedEntityAttributeSet.size());
			for(int i=0; i<mergedEntityAttributeSet.size(); i++) {
				List<String> list = mergedEntityAttributeSet.get(i);
				if(list.contains("Employee-id@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Employee-emp_id@Right"));
					assertTrue(list.contains("Employee-emp_id@Connector"));
				}
				else if(list.contains("Unit-name@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Department-name@Right"));
					assertTrue(list.contains("Department-name@Connector"));
				}
				else if(list.contains("Unit-head_count@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Department-employee_count@Right"));
					assertTrue(list.contains("Department-employee_count@Connector"));
				}
				else if(list.contains("Project-title@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Project-name@Right"));
					assertTrue(list.contains("Project-name@Connector"));
				}
				else if(list.contains("Employee-date_of_birth@Left")
						|| list.contains("Employee-age@Left")
						|| list.contains("Employee-name@Left")
						|| list.contains("Employee-name^last_name@Left")
						|| list.contains("Employee-name^first_name@Left")
						|| list.contains("Dependent-name@Left")
						|| list.contains("Dependent-date_of_birth@Left")
						|| list.contains("Dependent-relationship@Left")
						|| list.contains("Department-locations@Right")
						|| list.contains("Project-size@Right")
						|| list.contains("Project-start_date@Right")
						|| list.contains("Module-name@Right")
						|| list.contains("Module-start_date@Right")) {
					assertEquals(1, list.size());
				}
				else {
					assertTrue(false); //Ideally it should never get here. If it gets here, means there is bug! 
				}
			}
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testMergeSet_RelationshipAttribute() {
		try {			
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedRelationshipAttributeSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramRelationshipAttributeMapping, connectorToRightDiagramRelationshipAttributeMapping, 
					relationshipAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP});
			method.setAccessible(false);
			
			assertEquals(2, mergedRelationshipAttributeSet.size());
			for(int i=0; i<mergedRelationshipAttributeSet.size(); i++) {
				List<String> list = mergedRelationshipAttributeSet.get(i);
				if(list.contains("Employee-belongs_to-Unit-start_date@Left")) {
					assertEquals(3, list.size());
					assertTrue(list.contains("Employee-works_for-Department-since@Right"));
					assertTrue(list.contains("Employee-works_for-Department-since@Connector"));
				}
				else if(list.contains("Employee-manages-Department-start_date@Right")) {
					assertEquals(1, list.size());
				}
				else {
					assertTrue(false); //Ideally it should never get here. If it gets here, means there is bug! 
				}
			}
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testYieldDistinctMergedSet_Entity() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntitySet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityMapping, connectorToRightDiagramEntityMapping, entityDisjointUnion, LevelEnum.ENTITY});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedEntitySet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedEntitySet});
			method.setAccessible(false);
			
			assertEquals(5, distinctMergedEntitySet.size());
			assertTrue(distinctMergedEntitySet.contains("Employee@Connector"));
			assertTrue(distinctMergedEntitySet.contains("Department@Connector"));
			assertTrue(distinctMergedEntitySet.contains("Project@Connector"));
			assertTrue(distinctMergedEntitySet.contains("Dependent@Left"));
			assertTrue(distinctMergedEntitySet.contains("Module@Right"));
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testYieldDistinctMergedSet_Relationship() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedRelationshipSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramRelationshipMapping, connectorToRightDiagramRelationshipMapping, 
					relationshipDisjointUnion, LevelEnum.RELTIONSHIP});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, 
					new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedRelationshipSet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedRelationshipSet});
			method.setAccessible(false);
			
			assertEquals(7, distinctMergedRelationshipSet.size());
			assertTrue(distinctMergedRelationshipSet.contains("Employee-works_for-Department@Connector"));
			assertTrue(distinctMergedRelationshipSet.contains("Department-owns-Project@Connector"));
			assertTrue(distinctMergedRelationshipSet.contains("Employee-works_on-Project@Left"));
			assertTrue(distinctMergedRelationshipSet.contains("Employee-dependent_of-Dependent@Left"));
			assertTrue(distinctMergedRelationshipSet.contains("Employee-supervises-Employee@Right"));
			assertTrue(distinctMergedRelationshipSet.contains("Employee-manages-Department@Right"));
			assertTrue(distinctMergedRelationshipSet.contains("Project-contains-Module@Right"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testYieldDistinctMergedSet_EntityAttribute() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntityAttributeSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityAttributeMapping, connectorToRightDiagramEntityAttributeMapping, 
					entityAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_ENTITY});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedEntityAttributeSet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedEntityAttributeSet});
			method.setAccessible(false);
			
			assertEquals(17, distinctMergedEntityAttributeSet.size());
			assertTrue(distinctMergedEntityAttributeSet.contains("Employee-emp_id@Connector"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Department-name@Connector"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Department-employee_count@Connector"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Project-name@Connector"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Employee-date_of_birth@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Employee-age@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Employee-name@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Employee-name^last_name@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Employee-name^first_name@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Dependent-name@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Dependent-date_of_birth@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Dependent-relationship@Left"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Department-locations@Right"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Project-size@Right"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Project-start_date@Right"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Module-name@Right"));
			assertTrue(distinctMergedEntityAttributeSet.contains("Module-start_date@Right"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testYieldDistinctMergedSet_RelationshipAttribute() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedRelationshipAttributeSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramRelationshipAttributeMapping, connectorToRightDiagramRelationshipAttributeMapping, 
					relationshipAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedRelationshipAttributeSet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedRelationshipAttributeSet});
			method.setAccessible(false);
			
			assertEquals(2, distinctMergedRelationshipAttributeSet.size());
			assertTrue(distinctMergedRelationshipAttributeSet.contains("Employee-works_for-Department-since@Connector"));
			assertTrue(distinctMergedRelationshipAttributeSet.contains("Employee-manages-Department-start_date@Right"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }					
	}
	
	
	public void testMergeGraphWithEntities() {		
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntitySet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityMapping, connectorToRightDiagramEntityMapping, entityDisjointUnion, LevelEnum.ENTITY});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedEntitySet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedEntitySet});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_GRAPH_WITH_ENTITIES, 
					new Class[]{List.class, ERDiagram.class, ERDiagram.class, ERDiagram.class, ERDiagram.class});
			method.setAccessible(true); 
			method.invoke(erdiagramMerge, new Object[]{distinctMergedEntitySet, leftERDiagram, rightERDiagram, connector, mergedERDiagram});
			method.setAccessible(false);
			
			assertEquals(5, mergedERDiagram.getEntities().size());
			List<String> entityList = new ArrayList<String>();
			for(Entity entity: mergedERDiagram.getEntities()) {
				entityList.add(entity.getName());
			}
			assertTrue(entityList.contains("Employee"));
			assertTrue(entityList.contains("Department"));
			assertTrue(entityList.contains("Project"));
			assertTrue(entityList.contains("Dependent"));
			assertTrue(entityList.contains("Module"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }							
	}
	
	
	public void testMergeGraphWithRelationships() {
		testMergeGraphWithEntities();
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntitySet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityMapping, connectorToRightDiagramEntityMapping, entityDisjointUnion, LevelEnum.ENTITY});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedEntitySet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedEntitySet});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedRelationshipSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramRelationshipMapping, connectorToRightDiagramRelationshipMapping, 
					relationshipDisjointUnion, LevelEnum.RELTIONSHIP});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, 
					new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedRelationshipSet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedRelationshipSet});
			method.setAccessible(false);
			
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_GRAPH_WITH_RELATIONSHIPS, 
					new Class[]{List.class, List.class, List.class, ERDiagram.class, ERDiagram.class, ERDiagram.class, ERDiagram.class});
			method.setAccessible(true); 
			method.invoke(erdiagramMerge, new Object[]{distinctMergedRelationshipSet, mergedEntitySet, distinctMergedEntitySet, 
					leftERDiagram, rightERDiagram, connector, mergedERDiagram});
			method.setAccessible(false);
			
			assertEquals(7, mergedERDiagram.getRelationships().size());
			List<String> relationshipList = new ArrayList<String>();
			for(Relationship relationship: mergedERDiagram.getRelationships()) {
				relationshipList.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName());
			}
			assertTrue(relationshipList.contains("Employee-works_for-Department"));
			assertTrue(relationshipList.contains("Department-owns-Project"));
			assertTrue(relationshipList.contains("Employee-works_on-Project"));
			assertTrue(relationshipList.contains("Employee-dependent_of-Dependent"));
			assertTrue(relationshipList.contains("Employee-supervises-Employee"));
			assertTrue(relationshipList.contains("Employee-manages-Department"));
			assertTrue(relationshipList.contains("Project-contains-Module"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }								
	}
	
	
	public void testMergeGraphWithEntityAttributes() {
		testMergeGraphWithRelationships();
		try {			
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedEntityAttributeSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramEntityAttributeMapping, connectorToRightDiagramEntityAttributeMapping, 
					entityAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_ENTITY});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedEntityAttributeSet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedEntityAttributeSet});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_GRAPH_WITH_ENTITY_ATTRIBUTES, 
					new Class[]{List.class, List.class, ERDiagram.class, ERDiagram.class, ERDiagram.class, ERDiagram.class});
			method.setAccessible(true); 
			method.invoke(erdiagramMerge, new Object[]{distinctMergedEntityAttributeSet, mergedEntityAttributeSet, 
					leftERDiagram, rightERDiagram, connector, mergedERDiagram});
			method.setAccessible(false);
			
			List<String> employeeAttributeNames = new ArrayList<String>();
			List<String> departmentAttributeNames = new ArrayList<String>();
			List<String> projectAttributeNames = new ArrayList<String>();
			List<String> dependentAttributeNames = new ArrayList<String>();
			List<String> moduleAttributeNames = new ArrayList<String>();
			
			for(Entity entity: mergedERDiagram.getEntities()) {
				if("Employee".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes()) {
						employeeAttributeNames.add(attribute.getName());
						if(attribute.getName().equals("name")) {
							employeeAttributeNames.add("name^" + attribute.getAttributeAttributes().get(0).getName());
							employeeAttributeNames.add("name^" + attribute.getAttributeAttributes().get(1).getName());
						}
					}
				}
				else if("Department".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						departmentAttributeNames.add(attribute.getName());
				}
				else if("Project".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						projectAttributeNames.add(attribute.getName());
				}
				else if("Dependent".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						dependentAttributeNames.add(attribute.getName());
				}
				else if("Module".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						moduleAttributeNames.add(attribute.getName());
				}
				else {
					assertTrue(false); //Ideally it should never get here. If it gets here, means there is bug!
				}
			}
			
			assertTrue(employeeAttributeNames.contains("emp_id"));
			assertTrue(employeeAttributeNames.contains("age"));
			assertTrue(employeeAttributeNames.contains("date_of_birth"));
			assertTrue(employeeAttributeNames.contains("name"));
			assertTrue(employeeAttributeNames.contains("name^first_name"));
			assertTrue(employeeAttributeNames.contains("name^last_name"));
			
			assertTrue(departmentAttributeNames.contains("name"));
			assertTrue(departmentAttributeNames.contains("employee_count"));
			assertTrue(departmentAttributeNames.contains("locations"));
			
			assertTrue(projectAttributeNames.contains("size"));
			assertTrue(projectAttributeNames.contains("name"));
			assertTrue(projectAttributeNames.contains("start_date"));
			
			assertTrue(dependentAttributeNames.contains("name"));
			assertTrue(dependentAttributeNames.contains("date_of_birth"));
			assertTrue(dependentAttributeNames.contains("relationship"));
			
			assertTrue(moduleAttributeNames.contains("name"));
			assertTrue(moduleAttributeNames.contains("start_date"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }								
	}
	
	
	public void testMergeGraphWithRelationshipAttributes() {
		testMergeGraphWithRelationships();
		try {			
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_SET, 
					new Class[]{Map.class, Map.class, List.class, LevelEnum.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<List<String>> mergedRelationshipAttributeSet = (List<List<String>>)method.invoke(erdiagramMerge, 
					new Object[]{connectorToLeftDiagramRelationshipAttributeMapping, connectorToRightDiagramRelationshipAttributeMapping, 
					relationshipAttributeDisjointUnion, LevelEnum.ATTRIBUTE_OF_RELATIONSHIP});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_YIELD_DISTINCT_MERGED_SET, new Class[]{List.class});
			method.setAccessible(true); 
			@SuppressWarnings("unchecked")
			List<String> distinctMergedRelationshipAttributeSet = (List<String>)method.invoke(erdiagramMerge, new Object[]{mergedRelationshipAttributeSet});
			method.setAccessible(false);
			
			method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE_GRAPH_WITH_RELATIONSHIP_ATTRIBUTES, 
					new Class[]{List.class, List.class, ERDiagram.class, ERDiagram.class, ERDiagram.class, ERDiagram.class});
			method.setAccessible(true); 
			method.invoke(erdiagramMerge, new Object[]{distinctMergedRelationshipAttributeSet, mergedRelationshipAttributeSet, 
					leftERDiagram, rightERDiagram, connector, mergedERDiagram});
			method.setAccessible(false);
			
			List<String> managesAttributeNames = new ArrayList<String>();
			List<String> worksForAttributeNames = new ArrayList<String>();
			
			for(Relationship relationship: mergedERDiagram.getRelationships()) {
				if("manages".equals(relationship.getName())) {
					for(Attribute a: relationship.getRelationshipAttributes()) 
						managesAttributeNames.add(a.getName());
				}
				else if("works_for".equals(relationship.getName())) {
					for(Attribute a: relationship.getRelationshipAttributes()) 
						worksForAttributeNames.add(a.getName());
				}
			}
			
			assertEquals(1, managesAttributeNames.size());
			assertTrue(managesAttributeNames.contains("start_date"));
			assertEquals(1, worksForAttributeNames.size());
			assertTrue(worksForAttributeNames.contains("since"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }								
	}
	
	
	public void testMerge() {
		try {
			Method method = erdiagramMerge.getClass().getDeclaredMethod(METHOD_MERGE, new Class[]{ERDiagram.class, ERDiagram.class, Map.class});
			method.setAccessible(true); 
			ERDiagram mergedERDiagram = (ERDiagram)method.invoke(erdiagramMerge, new Object[]{leftERDiagram, rightERDiagram, leftToRightDiagramMapping});
			method.setAccessible(false);
			
			assertEquals(5, mergedERDiagram.getEntities().size());
			List<String> entityList = new ArrayList<String>();
			for(Entity entity: mergedERDiagram.getEntities()) {
				entityList.add(entity.getName());
			}
			assertTrue(entityList.contains("Employee"));
			assertTrue(entityList.contains("Department"));
			assertTrue(entityList.contains("Project"));
			assertTrue(entityList.contains("Dependent"));
			assertTrue(entityList.contains("Module"));
			
			assertEquals(7, mergedERDiagram.getRelationships().size());
			List<String> relationshipList = new ArrayList<String>();
			for(Relationship relationship: mergedERDiagram.getRelationships()) {
				relationshipList.add(relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName());
			}
			assertTrue(relationshipList.contains("Employee-works_for-Department"));
			assertTrue(relationshipList.contains("Department-owns-Project"));
			assertTrue(relationshipList.contains("Employee-works_on-Project"));
			assertTrue(relationshipList.contains("Employee-dependent_of-Dependent"));
			assertTrue(relationshipList.contains("Employee-supervises-Employee"));
			assertTrue(relationshipList.contains("Employee-manages-Department"));
			assertTrue(relationshipList.contains("Project-contains-Module"));
			
			List<String> employeeAttributeNames = new ArrayList<String>();
			List<String> departmentAttributeNames = new ArrayList<String>();
			List<String> projectAttributeNames = new ArrayList<String>();
			List<String> dependentAttributeNames = new ArrayList<String>();
			List<String> moduleAttributeNames = new ArrayList<String>();
			
			for(Entity entity: mergedERDiagram.getEntities()) {
				if("Employee".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes()) {
						employeeAttributeNames.add(attribute.getName());
						if(attribute.getName().equals("name")) {
							employeeAttributeNames.add("name^" + attribute.getAttributeAttributes().get(0).getName());
							employeeAttributeNames.add("name^" + attribute.getAttributeAttributes().get(1).getName());
						}
					}
				}
				else if("Department".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						departmentAttributeNames.add(attribute.getName());
				}
				else if("Project".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						projectAttributeNames.add(attribute.getName());
				}
				else if("Dependent".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						dependentAttributeNames.add(attribute.getName());
				}
				else if("Module".equals(entity.getName())) {
					for(Attribute attribute: entity.getEntityAttributes())
						moduleAttributeNames.add(attribute.getName());
				}
				else {
					assertTrue(false); //Ideally it should never get here. If it gets here, means there is bug!
				}
			}
			
			assertTrue(employeeAttributeNames.contains("emp_id"));
			assertTrue(employeeAttributeNames.contains("age"));
			assertTrue(employeeAttributeNames.contains("date_of_birth"));
			assertTrue(employeeAttributeNames.contains("name"));
			assertTrue(employeeAttributeNames.contains("name^first_name"));
			assertTrue(employeeAttributeNames.contains("name^last_name"));
			
			assertTrue(departmentAttributeNames.contains("name"));
			assertTrue(departmentAttributeNames.contains("employee_count"));
			assertTrue(departmentAttributeNames.contains("locations"));
			
			assertTrue(projectAttributeNames.contains("size"));
			assertTrue(projectAttributeNames.contains("name"));
			assertTrue(projectAttributeNames.contains("start_date"));
			
			assertTrue(dependentAttributeNames.contains("name"));
			assertTrue(dependentAttributeNames.contains("date_of_birth"));
			assertTrue(dependentAttributeNames.contains("relationship"));
			
			assertTrue(moduleAttributeNames.contains("name"));
			assertTrue(moduleAttributeNames.contains("start_date"));
			
			List<String> managesAttributeNames = new ArrayList<String>();
			List<String> worksForAttributeNames = new ArrayList<String>();
			
			for(Relationship relationship: mergedERDiagram.getRelationships()) {
				if("manages".equals(relationship.getName())) {
					for(Attribute a: relationship.getRelationshipAttributes()) 
						managesAttributeNames.add(a.getName());
				}
				else if("works_for".equals(relationship.getName())) {
					for(Attribute a: relationship.getRelationshipAttributes()) 
						worksForAttributeNames.add(a.getName());
				}
			}
			
			assertEquals(1, managesAttributeNames.size());
			assertTrue(managesAttributeNames.contains("start_date"));
			assertEquals(1, worksForAttributeNames.size());
			assertTrue(worksForAttributeNames.contains("since"));
			
		} catch (IllegalArgumentException e) {
		    e.printStackTrace();
		   } catch (NoSuchMethodException e) {
		    e.printStackTrace();
		   } catch (IllegalAccessException e) {
		    e.printStackTrace();
		   } catch (InvocationTargetException e) {
		    e.printStackTrace();
		   }			
	}
	
}
