/*
 * Copyright (c) 2012 Marsha Chechik, Alessio Di Sandro, Michalis Famelis,
 * Rick Salay, Vivien Suen.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    Wentao He - Implementation.
 */
package edu.toronto.cs.se.modelepedia.erdiagram.util;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import edu.toronto.cs.se.modelepedia.erdiagram.Attribute;
import edu.toronto.cs.se.modelepedia.erdiagram.Cardinality;
import edu.toronto.cs.se.modelepedia.erdiagram.ERDiagram;
import edu.toronto.cs.se.modelepedia.erdiagram.Entity;
import edu.toronto.cs.se.modelepedia.erdiagram.ErdiagramFactory;
import edu.toronto.cs.se.modelepedia.erdiagram.Relationship;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.ERDiagramTestDataGenerator;
import edu.toronto.cs.se.modelepedia.erdiagram.operator.util.ERDiagramMergeUtil;
import junit.framework.TestCase;

/**
 * Junit test cases for methods in ERDiagramMergeUtil.java
 */
public class TestERDiagramMergeUtil extends TestCase {
	
	private static final String DASH = "-";
	
	ERDiagram erdiagram;
	Map<String, String> diagramMapping;
	
	
	protected void setUp() throws Exception {
		erdiagram = ERDiagramTestDataGenerator.prepareLeftERDiagram();
		diagramMapping = ERDiagramTestDataGenerator.prepareLeftToRightDiagramMapping();
	}
	
	
	public void testLookupEntity() {
		Entity entity = erdiagram.getEntities().get(0);
		Entity lookedupEntity = ERDiagramMergeUtil.lookupEntity(erdiagram, entity.getName());
		
		assertEquals(entity, lookedupEntity);
	}

	
	public void testLookupRelationship() {
		Relationship relationship = erdiagram.getRelationships().get(0);
		Relationship lookedupRelationship = ERDiagramMergeUtil.lookupRelationship(erdiagram, 
				relationship.getFrom().getName() + DASH + relationship.getName() + DASH + relationship.getTo().getName());
		
		assertEquals(relationship, lookedupRelationship);
	}
	
	
	public void testLookupEntityAttribute() {
		Attribute attribute1 = ERDiagramMergeUtil.lookupEntityAttribute(erdiagram, "Employee-name^last_name");
		Attribute attribute2 = ERDiagramMergeUtil.lookupEntityAttribute(erdiagram, "Employee-name");
		Attribute attribute3 = ERDiagramMergeUtil.lookupEntityAttribute(erdiagram, "Employee-id");
		
		assertEquals("last_name", attribute1.getName());
		assertEquals("name", attribute2.getName());
		assertEquals("id", attribute3.getName());
	}
	
	
	public void testLookupRelationAttribute() {
		Attribute attribute = ERDiagramMergeUtil.lookupRelationshipAttribute(erdiagram, "Employee-belongs_to-Unit-start_date");
		
		assertEquals("start_date", attribute.getName());
	}
	
	
	public void testLookupMergedName() {
		List<String> distinctMergedSet = new ArrayList<String>();
		distinctMergedSet.add("Project@Left");
		distinctMergedSet.add("Employee@Left");
		distinctMergedSet.add("Department@Connector");
		
		List<List<String>> mergedSet = new ArrayList<List<String>>();
		List<String> list1 = new ArrayList<String>();
		list1.add("Unit@Left");
		list1.add("Department@Right");
		list1.add("Department@Connector");
		List<String> list2 = new ArrayList<String>();
		list2.add("Project@Left");
		List<String> list3 = new ArrayList<String>();
		list3.add("Employee@Left");
		mergedSet.add(list1);
		mergedSet.add(list2);
		mergedSet.add(list3);
		
		String mergedName1 = ERDiagramMergeUtil.lookupMergedName(distinctMergedSet, mergedSet, "Unit@Left");
		String mergedName2 = ERDiagramMergeUtil.lookupMergedName(distinctMergedSet, mergedSet, "Project@Left");
		
		assertEquals(mergedName1, "Department@Connector");
		assertEquals(mergedName2, "Project@Left");
	}
	
	
	public void testCloneEntity() {
		Entity entity = ErdiagramFactory.eINSTANCE.createEntity();
		entity.setName("Project");
		entity.setWeak(true);
		
		Entity clonedEntity = ERDiagramMergeUtil.cloneEntity(entity);
		
		assertEquals("Project", clonedEntity.getName());
		assertTrue(clonedEntity.isWeak());
	}
	
	
	public void testCloneRelationship() {
		Relationship relationship = ErdiagramFactory.eINSTANCE.createRelationship();
		relationship.setCardinality(Cardinality.MANY_TO_ONE);
		relationship.setName("has");
		relationship.setWeak(true);
		
		Relationship clonedRelationship = ERDiagramMergeUtil.cloneRelationship(relationship);
		
		assertEquals(Cardinality.MANY_TO_ONE, clonedRelationship.getCardinality());
		assertEquals("has", clonedRelationship.getName());
		assertTrue(clonedRelationship.isWeak());
	}
	
	
	public void testCloneAttribute() {
		Attribute attribute = ErdiagramFactory.eINSTANCE.createAttribute();
		attribute.setComposite(true);
		attribute.setDerived(false);
		attribute.setKey(true);
		attribute.setMultivalued(false);
		attribute.setName("id");
		
		Attribute clonedAttribute = ERDiagramMergeUtil.cloneAttribute(attribute);
		
		assertTrue(clonedAttribute.isComposite());
		assertFalse(clonedAttribute.isDerived());
		assertTrue(clonedAttribute.isKey());
		assertFalse(clonedAttribute.isMultivalued());
		assertEquals("id", clonedAttribute.getName());
	}
	
	
	public void testValidateEquality_Entity() {
		String message = "";
		Entity entity1 = ErdiagramFactory.eINSTANCE.createEntity();
		entity1.setName("Project");
		entity1.setWeak(true);
		Entity entity2 = ErdiagramFactory.eINSTANCE.createEntity();
		entity2.setName("Project");
		entity2.setWeak(false);
		
		try{
			ERDiagramMergeUtil.validateEquality(entity1, entity2);
		}
		catch(Exception e) {
			message = e.getMessage();
		}
		
		assertEquals(message, "Left entity 'Project' and right entity 'Project' have different properties.\n");
	}
	
	public void testValidateEquality_Relationship() {
		String message = "";
		Relationship relationship1 = ErdiagramFactory.eINSTANCE.createRelationship();
		relationship1.setCardinality(Cardinality.ONE_TO_MANY);
		relationship1.setName("has");
		relationship1.setWeak(true);
		Relationship relationship2 = ErdiagramFactory.eINSTANCE.createRelationship();
		relationship2.setCardinality(Cardinality.MANY_TO_ONE);
		relationship2.setName("owns");
		relationship2.setWeak(true);
		
		try{
			ERDiagramMergeUtil.validateEquality(relationship1, relationship2);
		}
		catch(Exception e) {
			message = e.getMessage();
		}
		
		assertEquals(message, "Left relationship 'has' and right relationship 'owns' have different properties.\n");
	}
	
	
	public void testValidateEquality_EntityAttribute() {
		String message = "";
		Entity entity1 = ErdiagramFactory.eINSTANCE.createEntity();
		entity1.setName("Project");
		entity1.setWeak(true);
		Entity entity2 = ErdiagramFactory.eINSTANCE.createEntity();
		entity2.setName("Project");
		entity2.setWeak(true);
		Attribute attribute1 = ErdiagramFactory.eINSTANCE.createAttribute();
		attribute1.setComposite(true);
		attribute1.setDerived(false);
		attribute1.setKey(true);
		attribute1.setMultivalued(false);
		attribute1.setName("id");
		Attribute attribute2 = ErdiagramFactory.eINSTANCE.createAttribute();
		attribute2.setComposite(true);
		attribute2.setDerived(false);
		attribute2.setKey(true);
		attribute2.setMultivalued(true);
		attribute2.setName("emp_id");
		
		try{
			ERDiagramMergeUtil.validateEquality(entity1, attribute1, entity2, attribute2);
		}
		catch(Exception e) {
			message = e.getMessage();
		}
		
		assertEquals(message, "Left attribute 'id' of entity 'Project' and right attribute 'emp_id' of entity 'Project' have different properties.\n");
	}
	
	
	public void testValidateEquality_RelationshipAttribute() {
		String message = "";
		Relationship relationship1 = ErdiagramFactory.eINSTANCE.createRelationship();
		relationship1.setCardinality(Cardinality.ONE_TO_MANY);
		relationship1.setName("has");
		relationship1.setWeak(true);
		Relationship relationship2 = ErdiagramFactory.eINSTANCE.createRelationship();
		relationship2.setCardinality(Cardinality.MANY_TO_ONE);
		relationship2.setName("owns");
		relationship2.setWeak(true);
		Attribute attribute1 = ErdiagramFactory.eINSTANCE.createAttribute();
		attribute1.setComposite(true);
		attribute1.setDerived(false);
		attribute1.setKey(true);
		attribute1.setMultivalued(false);
		attribute1.setName("id");
		Attribute attribute2 = ErdiagramFactory.eINSTANCE.createAttribute();
		attribute2.setComposite(true);
		attribute2.setDerived(false);
		attribute2.setKey(true);
		attribute2.setMultivalued(true);
		attribute2.setName("emp_id");
		
		try{
			ERDiagramMergeUtil.validateEquality(relationship1, attribute1, relationship2, attribute2);
		}
		catch(Exception e) {
			message = e.getMessage();
		}
		
		assertEquals(message, "Left attribute 'id' of relationship 'has' and right attribute 'emp_id' of relationship 'owns' have different properties.\n");
	}
	
	public void testGetEntityMapping() {
		Map<String, String> mapping = ERDiagramMergeUtil.getEntityMapping(diagramMapping);
		
		assertEquals(mapping.get("ENTITY@Project"), "ENTITY@Project");
		assertEquals(mapping.get("ENTITY@Unit"), "ENTITY@Department");
		assertEquals(mapping.get("ENTITY@Employee"), "ENTITY@Employee");
	}
	
	
	public void testGetRelationshipMapping() {
		Map<String, String> mapping = ERDiagramMergeUtil.getRelationshipMapping(diagramMapping);
		
		assertEquals(mapping.get("RELATIONSHIP@Unit-has-Project"), "RELATIONSHIP@Department-owns-Project");
		assertEquals(mapping.get("RELATIONSHIP@Employee-belongs_to-Unit"), "RELATIONSHIP@Employee-works_for-Department");
	}
	
	
	public void testGetEntityAttributeMapping() {
		Map<String, String> mapping = ERDiagramMergeUtil.getEntityAttributeMapping(diagramMapping);
		
		assertEquals(mapping.get("ENTITY_ATTRIBUTE@Project-title"), "ENTITY_ATTRIBUTE@Project-name");
		assertEquals(mapping.get("ENTITY_ATTRIBUTE@Employee-id"), "ENTITY_ATTRIBUTE@Employee-emp_id");
		assertEquals(mapping.get("ENTITY_ATTRIBUTE@Unit-head_count"), "ENTITY_ATTRIBUTE@Department-employee_count");
		assertEquals(mapping.get("ENTITY_ATTRIBUTE@Unit-name"), "ENTITY_ATTRIBUTE@Department-name");
	}
	
	
	public void testGetRelationshipAttributeMapping() {
		Map<String, String> mapping = ERDiagramMergeUtil.getRelationshipAttributeMapping(diagramMapping);
		
		assertEquals(mapping.get("RELATIONSHIP_ATTRIBUTE@Employee-belongs_to-Unit-start_date"), "RELATIONSHIP_ATTRIBUTE@Employee-works_for-Department-since");
	}
	
	
	public void testConvertToListOfList() {
		List<String> list = new ArrayList<String>();
		list.add("Project@Left");
		list.add("Employee@Left");
		list.add("Department@Connector");
		
		List<List<String>> listOfList = ERDiagramMergeUtil.convertToListOfList(list);
		
		for(List<String> li: listOfList) {
			assertEquals(1, li.size());
			assertTrue(li.contains("Project@Left")
					|| li.contains("Employee@Left")
					|| li.contains("Department@Connector"));
		}
	}
	
	
	public void testMergeList() {
		List<List<String>> listOfList = new ArrayList<List<String>>();
		List<String> list1 = new ArrayList<String>();
		list1.add("Unit@Left");
		list1.add("Department@Right");
		list1.add("Department@Connector");
		List<String> list2 = new ArrayList<String>();
		list2.add("Project@Left");
		List<String> list3 = new ArrayList<String>();
		list3.add("Employee@Left");
		listOfList.add(list1);
		listOfList.add(list2);
		listOfList.add(list3);
		
		ERDiagramMergeUtil.mergeList(listOfList, "Unit@Left", "Employee@Left");
		
		assertEquals(2, listOfList.size());
		assertEquals(4, listOfList.get(0).size());
		assertTrue(listOfList.get(0).contains("Unit@Left"));
		assertTrue(listOfList.get(0).contains("Department@Right"));
		assertTrue(listOfList.get(0).contains("Department@Connector"));
		assertTrue(listOfList.get(0).contains("Employee@Left"));
		assertEquals(1, listOfList.get(1).size());
		assertTrue(listOfList.get(1).contains("Project@Left"));
	}
	
}
